#!/usr/bin/python3.5
#Version 34
import requests
from requests.packages.urllib3.exceptions import InsecureRequestWarning
import xml.etree.ElementTree as ET
import json

import sys
import argparse
import os

# A subdirectory for verbose output and files
path='debug' 

# Used for logon and HMC comms authority
HMCname = ''
token = ''
tokenFilename = "HMC.token"

#----- Functions

def save_to_file(filename, string):
   if debug:
     filename = path + '/' + filename
     file = open(filename,"w")
     file.write(string)
     file.close()
     print("------- Saved data to file " + filename)


def HMClogon(user, passwd, reuse):
    logonheaders = {'Content-Type': 'application/vnd.ibm.powervm.web+xml; type=LogonRequest'}
    logonUrl =  'https://'+HMCname+':12443/rest/api/web/Logon'
    logonPayload = '<LogonRequest schemaVersion="V1_0" xmlns="http://www.ibm.com/xmlns/systems/power/firmware/web/mc/2012_10/" xmlns:mc="http://www.ibm.com/xmlns/systems/power/firmware/web/mc/2012_10/"><UserID>' + user + '</UserID><Password>' + passwd + '</Password></LogonRequest>'
    ret = requests.put(logonUrl,data=logonPayload,headers=logonheaders,verify=False)
    if ret.status_code != 200:
        print("Error: Logon failed error code=%d url=%s" %(ret.status_code,logonUrl))
        if debug:
                print("Returned:" + ret.text)
        # Do not return if we failed to logon
        sys.exit(ret.status_code)
    #Extract from the XML the token
    xmlResponse = ET.fromstring(ret.text)
    token = xmlResponse[1].text
    if debug:
        print("Log on response=%d and got Token=\n----\n%s\n----" %(ret.status_code, token))
    if reuse:
        print("Save session token in local directory file:%s" %(tokenFilename))
        file_object = open(tokenFilename,"w")
        file_object.write(token)
        file_object.close()
    return token


def HMClogoff():
    if reuse:
        print("Leave %s file for next time" %(tokenFilename))
        sys.exit(0)
    logoffheaders = {'X-API-Session' : token }
    logoffUrl =  'https://'+HMCname+':12443/rest/api/web/Logon'
    ret = requests.delete(logoffUrl,headers=logoffheaders,verify=False)
    rcode = ret.status_code
    # delete can respond with these three good values
    if rcode == 200 or rcode == 202 or rcode == 204:
        print("Successfully disconnected from the HMC (code=%d)" %(rcode))
        if os.path.isfile(tokenFilename):
            os.remove(tokenFilename)
        sys.exit(0)
    else:
        print("Error: Logoff failed error code=%d url=%s data=%s" %(rcode, logoffUrl, ret.text))
        sys.exit(rcode)


def get_preferences():
    print("------- Get Preferences")
    prefHeaders = {'X-API-Session' : token}
    prefUrl = 'https://'+HMCname+':12443/rest/api/pcm/preferences'
    ret = requests.get(prefUrl, headers=prefHeaders,verify=False)
    if ret.status_code != 200:
        print("Error: GET preferences error code=%d header=%s url=%s" %(ret.status_code, prefHeaders, prefUrl))
        if ret.status_code == 401:
            print("Unauthorisied either wrong User/Password wrong or HMC.token out of date")
        if debug:
            print("Returned: data=%s" %( ret.text))
        HMClogoff()
    save_to_file('energy-prefs.xml', ret.text)
    return ret.text

# the preferences returned by the HMC contant at the start
#  <feed> 
#    .....
#    <entry>
#        .....
#        <content> 
#        <ManagementConsole= HMC> 
#            <ManagementSystem = Server> 
#              server details
#            <ManagementSystem> 
#            ... 
#        </ManagementConsole= HMC> 
#      </content> 
#    </entry> 
#  </feed> 
# And the closing of those sections at the end which MUST NOT BE returned to the HMC to set the perfernence
# Here they are stripped off using string manipulation

def strip_preferences(pref_string):
    #Find the first part of what we need then remove everything before the offset
    i = pref_string.find("<ManagementConsolePcmPreference:ManagementConsolePcmPreference")
    pref_string = pref_string[i : ]
    # rip off the trailer
    #Find the first part of what we don't need at the end then remove everything after the offset
    i = pref_string.find("</content>")
    pref_string = pref_string[ :i ]
    return pref_string


# To get Energy stats we need for the Server
# 1. EnergyMonitoringCapable = 'true' - read-only set by FSP
# 2. AggregationEnabled = 'true' Setable, generates ProcessesMetrics (which we are using)
# 3. EnergyMonitorEnabled = 'true' Setable. Can't switch off if AggregationEnabled = 'true'
def check_preferences(server, get_pref_result):
    print("------- Check if we need to set Preferences")
    save_to_file('energy-prefs-stipped.xml', get_pref_result)
    # Convert string to XML
    root = ET.fromstring(get_pref_result)
    if debug:
        tree = ET.ElementTree(root)
        #string = tree.write(file_name)
        filename = path + 'energy-prefs-before_updated.xml'
        tree.write(filename)
    AtomID = 'none'
    result = 'not-found'

    # Set the namespace
    ns = {'yyy': 'http://www.ibm.com/xmlns/systems/power/firmware/pcm/mc/2012_10/'}

    for this in root.findall('yyy:ManagedSystemPcmPreference',ns):
        name      = this.find('yyy:SystemName',ns)
        aggregate = this.find('yyy:AggregationEnabled',ns)
        capable   = this.find('yyy:EnergyMonitoringCapable',ns)
        enabled   = this.find('yyy:EnergyMonitorEnabled',ns)
        print('FOUND System=%32s Aggregation=%6s Capable=%6s Enabled=%6s' %(name.text, aggregate.text, capable.text, enabled.text))

        meta   = this.find('yyy:Metadata',ns)
        atom = meta.find('yyy:Atom',ns)
        atomitem = atom.find('yyy:AtomID',ns)
        #print(atomid.text)
        if server == name.text:
            result = 'found'
            AtomID = atomitem.text
            if capable.text == 'true':
                print("    Server can do Energy monitoring")
                if enabled.text == 'true' and aggregate.text == 'true':
                    print("    Aggregation and Energy monitoring already set")
                    result='already set'
                else:
                    print("    Aggregation or Energy monitoring needs setting")
                    result='needs-setting'
                    # Actually change this server to EnergyMonitorEnabled = 'true'
                    # ready for the Set Perferences
                    aggregate.text = 'true'
                    enabled.text = 'true'
            else:
                print("    Server can't do Energy monitoring")
                result='impossible'
    return result, root, AtomID
 

def set_preferences(xmlPref):
    tree = ET.ElementTree(xmlPref)
    if debug:
        filename= path + 'energy-prefs-updated.xml'
        tree.write(filename)

    postHeaders = {'Content-Type': 'application/xml', 'X-API-Session' : token }
    postUrl = 'https://'+HMCname+':12443/rest/api/pcm/preferences'
    xmlStr = ET.tostring(xmlPref)
    r = requests.post(postUrl, headers=postHeaders, data=xmlStr, verify=False)
    if r.status_code != 200:
        print("Error: post preferences error code=%d header=%s url=%s" %( r.status_code,postHeaders,postUrl))
        if debug:
            print("URL:" + postUrl)
            print("Body:")
            print(xmlStr)
            print("Returned:" + r.text)
        HMClogoff()


def get_filenames(Atomid, Server):
  print("------- Get JSON filenames")
  fileHeaders = {'X-API-Session' : token, 'Accept': 'application/atom+xml' }
  fileUrl =  'https://'+HMCname+':12443/rest/api/pcm/ManagedSystem/'+Atomid+'/ProcessedMetrics?Type=Energy' 
  r = requests.get(fileUrl,headers=fileHeaders,verify=False)
  if r.status_code != 200:
        print("Error %d: Get Server %s file of filenames \n\tHeader=%s \n\tURL=%s" \
              %( r.status_code,Server,fileHeaders,fileUrl))
        if r.status_code == 204:
            print("Hint: 204=No Content - Have you updated the Preferences to switch on monitoring? See --prefs option")
            print("Hint: 204=No Content - Wait 6 minutes after you switch on stats to get some data")
        if debug:
            print("Returned:" + r.text)
  else:
    if debug:
        file_name = "Server-filenames-" + Server + ".xml"
        save_to_file(file_name, r.text)
  return r.status_code, r.text


def get_stats(JSON_url, JSON_file, Server):
  jsonHeaders = {'X-API-Session' : token }
  r = requests.get(JSON_url,headers=jsonHeaders,verify=False)
  if r.status_code != 200:
    print("Error: get Server JSON file error code=%d filename=%s" %( r.status_code, JSON_file))
    if debug:
            print("Returned:" + r.text)
    return ""
  if debug:
    #file_name = "Server-" + Server + "-Processed-Energy.json" + JSON_file
    file_name = Server + "-Processed-Energy.json"
    save_to_file(file_name, r.text)
  return r.text


# ----- MAIN CODE
# define the command line arguments
parser = argparse.ArgumentParser()
parser.add_argument("--hmc", help="Specify the HMC IP Address or Hostname")
parser.add_argument("--username", help="username for the HMC")
parser.add_argument("--password", help="password for the HMC")
parser.add_argument("--server", help="Server name")
parser.add_argument("--debug", help="Output details as we go + files in directory debug",action="store_true")
parser.add_argument("--reuse", help="Save and reuse HMC token",action="store_true")
# extract the command line arguments
args = parser.parse_args()
HMCname = args.hmc
userName = args.username
serverName = args.server
password = args.password
debug = args.debug
reuse = args.reuse
# check the command line arguments
if(HMCname == None):
    sys.exit('HMC --hmc <IP or host name> is mandatory')
if(userName == None):
    sys.exit('HMC --username is mandatory')
if(password == None):
    sys.exit('HMC --password is mandatory')
if(serverName == None):
    sys.exit('HMC --Server is mandatory')

# if --debug then output goes to a sub-diectory with the below name 
if debug:
    if os.path.isdir(path) == False:
        print('You need a sub-directory called "debug" for the --debug output files')
        sys.exit(42)
# if --reuse then we save the token for HMC access this is the file used

if debug:
    print("Switching off ugly Security Warnings 'Unverified HTTPS request is being made'")

requests.packages.urllib3.disable_warnings(InsecureRequestWarning)

print("------- Server name: %s   Debug:%d Reuse:%d (0=No)" %(serverName, debug, reuse))
# if --reuse load the token file and try it. It could be invalid or too old
response = ""
if reuse:
  if os.path.isfile(tokenFilename) and os.access(tokenFilename, os.R_OK):
    print("HMC.token exists and is readable")
    file_object = open(tokenFilename,"r")
    token = file_object.read()
    file_object.close()
    # Test the token by extracting the preferences
    response = get_preferences()
    if response == "":
        print("HMC.token file did not work")
        os.remove(tokenFilename)
    else:
        print("HMC.token file worked")
  else:
    print("No readable HMC.token file found")
# if --reuse failed or not --reuse try a full logon
if response == "":
  print("HMC Logon")
  token = HMClogon(userName, password, reuse)
  response = get_preferences()
# if we can't communicate with the HMC stop here
if response == "":
  print("HMC Login or Get server Preferences failed ... stopping")
  HMClogoff()

reducedResponse = strip_preferences(response)
status, fixedXml, AtomID = check_preferences(serverName,reducedResponse)
print('End of Servers')
if status == 'not-found':
  print('Server %s not found'%(serverName))
if status == 'impossible':
  print('Server %s does not support Energy Monitoring'%(serverName))
if status == 'not-found' or status == 'impossible':
  HMClogoff()

if status == 'needs-setting':
  set_preferences(fixedXml)
  print("------- Successfully updated Preferences")
else:
  print("------- Not Updating Preferences")


status, filenames_str= get_filenames(AtomID, serverName)
if status != 200:
  HMClogoff()

# Parse the file of filenames
filenames_xml = ET.fromstring(filenames_str)
for reffeed in filenames_xml:
    if 'entry' in reffeed.tag:
        for entry in reffeed:
            if 'title' in entry.tag:
                filename = entry.text
            if 'link' in entry.tag:
                JSON_file_url = entry.get('href')
        if debug:
                print("JSON filename |%s|" % (filename))
                print("JSON URL |%s|" % (JSON_file_url))
        statResponse = get_stats(JSON_file_url, filename, serverName)
        # convert to JSON
        if statResponse == "":
                    print("Getting JSON stats failed continuing ...")
                    continue

        data = json.loads(statResponse)
        # end and start TimeStamps went missing in a recent HMC upgrade.
        #endTimeStamp = data['systemUtil']['utilInfo']['endTimeStamp']
        #startTimeStamp = data['systemUtil']['utilInfo']['startTimeStamp']
        frequency  = data['systemUtil']['utilInfo']['frequency']
        name = data['systemUtil']['utilInfo']['name']
        mtms = data['systemUtil']['utilInfo']['mtms']
        i = mtms.find('*')
        mtm = mtms[ :i ]
        serial = mtms[i: ]
        mtms = mtms.replace('*', ',') # sepeate with a comma.
        print("Server %s MTMS=%s MTM=|%s| Serial=|%s|" %(name,mtms,mtm,serial))
        if debug :
            #print("Start=%s" %(startTimeStamp))
            #print("End  =%s" %(endTimeStamp))
            print("Frequency=%s seconds" %(frequency))

        file_name = "energy.csv"
        file_object = open(file_name,"a")
        count = 0
        errors = 0
        for sample in data['systemUtil']['utilSamples']:
            if sample['sampleInfo']['status'] != 0 :
               print("\tError Server %s: Sample status=%d (not zero) time=%s" %(name,sample['sampleInfo']['status'],
                                                 sample['sampleInfo']['timeStamp']))
               errors+= 1
            else :
                count += 1
                watts  = sample['energyUtil']['powerUtil']['powerReading'][0]
                mb0    = sample['energyUtil']['thermalUtil']['baseboardTemperatures'][0]['temperatureReading'][0]
                mb1    = sample['energyUtil']['thermalUtil']['baseboardTemperatures'][1]['temperatureReading'][0]
                mb2    = sample['energyUtil']['thermalUtil']['baseboardTemperatures'][2]['temperatureReading'][0]
                if mtm == '8408-E8E':
                    mb3    = sample['energyUtil']['thermalUtil']['baseboardTemperatures'][3]['temperatureReading'][0]

                cpu0    = sample['energyUtil']['thermalUtil']['cpuTemperatures'][0]['temperatureReading'][0]
                cpu1    = sample['energyUtil']['thermalUtil']['cpuTemperatures'][1]['temperatureReading'][0]
                cpu2    = sample['energyUtil']['thermalUtil']['cpuTemperatures'][2]['temperatureReading'][0]
                cpu3    = sample['energyUtil']['thermalUtil']['cpuTemperatures'][3]['temperatureReading'][0]
                if mtm == '8408-E8E':
                    cpu4    = sample['energyUtil']['thermalUtil']['cpuTemperatures'][4]['temperatureReading'][0]
                    cpu5    = sample['energyUtil']['thermalUtil']['cpuTemperatures'][5]['temperatureReading'][0]
                    cpu6    = sample['energyUtil']['thermalUtil']['cpuTemperatures'][6]['temperatureReading'][0]
                    cpu7    = sample['energyUtil']['thermalUtil']['cpuTemperatures'][7]['temperatureReading'][0]

                inlet   = sample['energyUtil']['thermalUtil']['inletTemperatures'][0]['temperatureReading'][0]
                timeStamp = sample['sampleInfo']['timeStamp']
                if mtm == '8408-E8E':
                    buffer = "%s,%s,%s-%02d-%s-%s-%s-%s,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d" % (name,mtms,timeStamp[0:4],int(timeStamp[5:7]) -1, timeStamp[8:10], timeStamp[11:13], timeStamp[14:16],timeStamp[17:19], watts, mb0,mb1,mb2,mb3,cpu0,cpu1,cpu2,cpu3,cpu4,cpu6,cpu6,cpu7,inlet)
                else:
                    buffer = "%s,%s,%s-%02d-%s-%s-%s-%s,%d,%d,%d,%d,%d,%d,%d,%d,%d" % (name,mtms,timeStamp[0:4],int(timeStamp[5:7]) -1, timeStamp[8:10], timeStamp[11:13], timeStamp[14:16],timeStamp[17:19], watts, mb0,mb1,mb2,cpu0,cpu1,cpu2,cpu3,inlet)
                if debug:
                    print(buffer)
                file_object.write(buffer + '\n')

        file_object.close()
        print("Processed: %d errors=%d" % (count,errors))

HMClogoff()
